from copy import deepcopy
import torch
from models import Discriminator, Generator  # Custom library for modeling

# Create a Discriminator (x) and a Generator (y)
netD = Discriminator(...)   # $x_0$
netG = Generator(...)       # $y_0$
netG_ = deepcopy(netG)      # $\tilde{y}_0$

# Define the optimizers
optimizerD = torch.optim.Adam(netD.parameters(), ...)
optimizerG = torch.optim.Adam(netG.parameters(), ...)

dataloader = ...    # set of real images
num_epochs = ...    # number of epochs
criterion  = ...    # loss function 
label_real = ...    # label for real image e.g. all ones
label_fake = ...    # label for fake image e.g. all zeros
gamma = ...     # Alex-GDA parameter
delta = ...     # Alex-GDA parameter

for epoch in range(1, num_epochs+1):
  for data in dataloader:
    # Generate latent vectors
    noise = torch.randn(...)

    # Save $x_k$ to `netD_`
    netD_ = deepcopy(netD)  

    # Compute Discriminator error: $f(x_k, \tilde{y}_k)$
    errD_real = criterion(netD(data), label_real) 
    errD_fake = criterion(netD(netG_(noise)), label_fake)   # `netG_` == $\tilde{y}_k$
    errD = errD_real + errD_fake 
    
    # Update Discriminator: $x_{k+1}$ 
    optimizerD.zero_grad()
    errD.backward()
    optimizerD.step()

    # Interpolation/Extrapolation step for x:
    # Compute $\tilde{x}_{k+1} = \gamma * x_{k+1} + (1-\gamma) * x_k$ and save to `netD_`
    for online, target in zip(netD.parameters(), netD_.parameters()):
      target.data = gamma * online.data + (1 - gamma) * target.data
    
    # Save $y_k$ to `netG_`
    netG_ = deepcopy(netG)

    # Compute Generator error: $f(\tilde{x}_{k+1}, y_k)$
    errG = criterion(netD_(netG(noise)), label_real)    # `netD_` == $\tilde{x}_{k+1}$

    # Update Generator: $y_{k+1}$ 
    optimizerG.zero_grad()
    errG.backward()
    optimizerG.step()

    # Interpolation/Extrapolation step for y:
    # Compute $\tilde{y}_{k+1} = \delta * y_{k+1} + (1-\delta) * y_k$ and save to `netG_`
    for online, target in zip(netG.parameters(), netG_.parameters()):
      target.data = delta * online.data + (1 - delta) * target.data